package net.knarcraft.ffmpegconverter.streams;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * A class for representing stream tags that might be found for streams in video files
 */
public enum StreamTag {

    /**
     * The absolute index of this stream in the file
     */
    INDEX("index", StreamType.ALL),
    /**
     * The name of the codec, useful for identification
     */
    CODEC_NAME("codec_name=", StreamType.ALL),

    /**
     * The long name of the codec, useful for displaying information
     */
    CODEC_LONG_NAME("codec_long_name", StreamType.ALL),

    /**
     * The profile the encoder is set to
     */
    PROFILE("profile", StreamType.ALL),

    /**
     * Whether the type of codec for the stream is audio, video or subtitle
     */
    CODEC_TYPE("codec_type", StreamType.ALL),
    CODEC_TAG_STRING("codec_tag_string", StreamType.ALL),
    CODEC_TAG("codec_tag", StreamType.ALL),
    SAMPLE_FMT("sample_fmt", Set.of(StreamType.AUDIO)),
    SAMPLE_RATE("sample_rate", Set.of(StreamType.AUDIO)),

    /**
     * The number of channels in an audio stream
     */
    CHANNELS("channels", Set.of(StreamType.AUDIO)),

    /**
     * Human-recognizable term for the number of audio channels, such as stereo, mono or surround
     */
    CHANNEL_LAYOUT("channel_layout", Set.of(StreamType.AUDIO)),
    BITS_PER_SAMPLE("bits_per_sample", Set.of(StreamType.AUDIO)),
    INITIAL_PADDING("initial_padding", Set.of(StreamType.AUDIO)),

    /**
     * The viewable video width
     */
    WIDTH("width", StreamType.VIDEO_SUBTITLE),

    /**
     * The viewable video height
     */
    HEIGHT("height", StreamType.VIDEO_SUBTITLE),

    /**
     * The original video width, before any padding was applied to account for resolution multiples
     */
    CODED_WIDTH("coded_width", Set.of(StreamType.VIDEO)),

    /**
     * The original video height, before any padding was applied to account for resolution multiples
     */
    CODED_HEIGHT("coded_height", Set.of(StreamType.VIDEO)),
    CLOSED_CAPTIONS("closed_captions", Set.of(StreamType.VIDEO)),
    FILM_GRAIN("film_grain", Set.of(StreamType.VIDEO)),
    HAS_B_FRAMES("has_b_frames", Set.of(StreamType.VIDEO)),

    /**
     * The aspect ratio used to stretch the video for playback
     */
    SAMPLE_ASPECT_RATIO("sample_aspect_ratio", Set.of(StreamType.VIDEO)),

    /**
     * The aspect ratio of the video stream
     */
    DISPLAY_ASPECT_RATIO("display_aspect_ratio", Set.of(StreamType.VIDEO)),

    /**
     * The pixel format used for the video stream
     */
    PIX_FMT("pix_fmt", Set.of(StreamType.VIDEO)),

    /**
     * The quality level of the video stream
     */
    LEVEL("level", Set.of(StreamType.VIDEO)),
    COLOR_RANGE("color_range", Set.of(StreamType.VIDEO)),

    /**
     * How colors are stored in the video stream's file
     */
    COLOR_SPACE("color_space", Set.of(StreamType.VIDEO)),
    COLOR_TRANSFER("color_transfer", Set.of(StreamType.VIDEO)),
    COLOR_PRIMARIES("color_primaries", Set.of(StreamType.VIDEO)),
    CHROMA_LOCATION("chroma_location", Set.of(StreamType.VIDEO)),
    FIELD_ORDER("field_order", Set.of(StreamType.VIDEO)),
    REFS("refs", Set.of(StreamType.VIDEO)),
    IS_AVC("is_avc", Set.of(StreamType.VIDEO)),
    NAL_LENGTH_SIZE("nal_length_size", Set.of(StreamType.VIDEO)),
    ID("id", StreamType.ALL),
    R_FRAME_RATE("r_frame_rate", StreamType.ALL),
    AVERAGE_FRAME_RATE("avg_frame_rate", StreamType.ALL),
    TIME_BASE("time_base", StreamType.ALL),
    START_PTS("start_pts", StreamType.ALL),
    START_TIME("start_time", StreamType.ALL),
    DURATION_TS("duration_ts", StreamType.ALL),
    DURATION("duration", StreamType.ALL),
    BIT_RATE("bit_rate", StreamType.ALL),
    MAX_BIT_RATE("max_bit_rate", StreamType.ALL),
    BITS_PER_RAW_SAMPLE("bits_per_raw_sample", StreamType.ALL),
    NB_FRAMES("nb_frames", StreamType.ALL),
    NB_READ_FRAMES("nb_read_frames", StreamType.ALL),
    NB_READ_PACKETS("nb_read_packets", StreamType.ALL),
    EXTRA_DATA_SIZE("extradata_size", StreamType.VIDEO_SUBTITLE),
    DISPOSITION_DEFAULT("DISPOSITION:default", StreamType.ALL),
    DISPOSITION_DUB("DISPOSITION:dub", StreamType.ALL),
    DISPOSITION_ORIGINAL("DISPOSITION:original", StreamType.ALL),
    DISPOSITION_COMMENT("DISPOSITION:comment", StreamType.ALL),
    DISPOSITION_LYRICS("DISPOSITION:lyrics", StreamType.ALL),
    DISPOSITION_KARAOKE("DISPOSITION:karaoke", StreamType.ALL),
    DISPOSITION_FORCED("DISPOSITION:forced", StreamType.ALL),
    DISPOSITION_HEARING_IMPAIRED("DISPOSITION:hearing_impaired", StreamType.ALL),
    DISPOSITION_VISUAL_IMPAIRED("DISPOSITION:visual_impaired", StreamType.ALL),
    DISPOSITION_CLEAN_EFFECTS("DISPOSITION:clean_effects", StreamType.ALL),
    DISPOSITION_ATTACHED_PIC("DISPOSITION:attached_pic", StreamType.ALL),
    DISPOSITION_TIMED_THUMBNAILS("DISPOSITION:timed_thumbnails", StreamType.ALL),
    DISPOSITION_NON_DIEGETIC("DISPOSITION:non_diegetic", StreamType.ALL),
    DISPOSITION_CAPTIONS("DISPOSITION:captions", StreamType.ALL),
    DISPOSITION_DESCRIPTIONS("DISPOSITION:descriptions", StreamType.ALL),
    DISPOSITION_METADATA("DISPOSITION:metadata", StreamType.ALL),
    DISPOSITION_DEPENDENT("DISPOSITION:dependent", StreamType.ALL),
    DISPOSITION_STILL_IMAGE("DISPOSITION:still_image", StreamType.ALL),

    /**
     * The language of the stream
     */
    TAG_LANGUAGE("TAG:language", StreamType.ALL),

    /**
     * The title of an audio stream
     */
    TAG_TITLE("TAG:title", StreamType.ALL),
    TAG_BPS_ENG("TAG:BPS-eng", StreamType.ALL),
    TAG_DURATION_ENG("TAG:DURATION-eng", StreamType.ALL),
    TAG_NUMBER_OF_FRAMES_ENG("TAG:NUMBER_OF_FRAMES-eng", StreamType.ALL),
    TAG_NUMBER_OF_BYTES_ENG("TAG:NUMBER_OF_BYTES-eng", StreamType.ALL),
    TAG_SOURCE_ID_ENG("TAG:SOURCE_ID-eng", StreamType.ALL),
    TAG_SOURCE_ID("TAG:SOURCE_ID", StreamType.ALL),
    TAG_STATISTICS_WRITING_APP_ENG("TAG:_STATISTICS_WRITING_APP-eng", StreamType.ALL),
    TAG_STATISTICS_WRITING_APP("TAG:_STATISTICS_WRITING_APP", StreamType.ALL),
    TAG_STATISTICS_WRITING_DATE_UTC_ENG("TAG:_STATISTICS_WRITING_DATE_UTC-eng", StreamType.ALL),
    TAG_STATISTICS_WRITING_DATE_UTC("TAG:_STATISTICS_WRITING_DATE_UTC", StreamType.ALL),
    TAG_STATISTICS_TAGS_ENG("TAG:_STATISTICS_TAGS-eng", StreamType.ALL),
    TAG_STATISTICS_TAGS("TAG:_STATISTICS_TAGS", StreamType.ALL),
    TAG_ENCODER("TAG:ENCODER", Set.of(StreamType.VIDEO)),
    TAG_DURATION("TAG:DURATION", StreamType.ALL),
    ;

    private static final Map<String, StreamTag> tagLookup = new HashMap<>();

    private final @NotNull String tagString;
    private final Set<StreamType> applicableFor;

    /**
     * Instantiates a new stream tag
     *
     * @param tagString <p>The tag string ffmpeg prints to specify this tag</p>
     */
    StreamTag(@NotNull String tagString, @NotNull Set<StreamType> applicableFor) {
        this.tagString = tagString;
        this.applicableFor = applicableFor;
    }

    /**
     * Gets the types of streams this tag is applicable for
     *
     * @return <p>The types of streams this tag is applicable for</p>
     */
    @NotNull
    public Set<StreamType> getApplicableFor() {
        return new HashSet<>(this.applicableFor);
    }

    /**
     * Gets the stream tag defined by the given string
     *
     * @param input <p>The input string to parse</p>
     * @return <p>The corresponding stream tab, or null if not found</p>
     */
    @Nullable
    public static StreamTag getFromString(@NotNull String input) {
        if (tagLookup.isEmpty()) {
            for (StreamTag tag : StreamTag.values()) {
                tagLookup.put(tag.tagString, tag);
            }
        }
        return tagLookup.get(input);
    }

    @Override
    public String toString() {
        return this.tagString;
    }

}
